/*
************************************************************************
*
*   ExSurface.c - AddSurface command
*
*   Copyright (c) 1994-95
*
*   ETH Zuerich
*   Institut fuer Molekularbiologie und Biophysik
*   ETH-Hoenggerberg
*   CH-8093 Zuerich
*
*   SPECTROSPIN AG
*   Industriestr. 26
*   CH-8117 Faellanden
*
*   All Rights Reserved
*
*   Date of last modification : 96/04/21
*   Pathname of SCCS file     : /sgiext/molmol/src/cmdprim/SCCS/s.ExSurface.c
*   SCCS identification       : 1.20
*
************************************************************************
*/

#include <cmd_prim.h>

#include <stdio.h>
#include <stdlib.h>
#include <values.h>
#include <math.h>

#include <break.h>
#include <surface.h>
#include <arg.h>
#include <par_names.h>
#include <par_hand.h>
#include <data_hand.h>
#include <data_sel.h>
#include <prim_hand.h>
#include <attr_struc.h>
#include <graph_draw.h>

static float SolventRad = 1.4;
static int SurfKind = 0;
static BOOL SurfDots = TRUE, SurfFast = TRUE;

static SphereDescr *SphereA;
static int SphereNo;

static void
countAtom(DhAtomP atomP, void *clientData)
{
  (* (int *) clientData)++;
}

static void
fillData(DhAtomP atomP, void *clientData)
{
  ExprP exprP = clientData;
  ExprRes exprRes;

  SphereA[SphereNo].userData = atomP;
  DhAtomGetCoord(atomP, SphereA[SphereNo].cent);

  ExprEval(atomP, exprP, &exprRes);
  if (exprRes.resType == ER_INT)
    SphereA[SphereNo].rad = (float) exprRes.u.intVal;
  else
    SphereA[SphereNo].rad = exprRes.u.floatVal;

  SphereNo++;
}

static void
addSurface(DhMolP molP, void *clientData)
{
  PropRefP refP;
  int atomNo;
  PatchDescr *patchA;
  int patchNo;
  IsoSurface *surfA;
  int surfNo, surfI;
  PrimObjP primP;
  int i;

  if (BreakInterrupted())
    return;

  refP = PropGetRef(PROP_SELECTED, FALSE);

  atomNo = 0;
  DhMolApplyAtom(refP, molP, countAtom, &atomNo);
  if (atomNo == 0)
    return;

  SphereA = malloc(atomNo * sizeof(*SphereA));
  SphereNo = 0;
  DhMolApplyAtom(refP, molP, fillData, clientData);

  if (SurfKind == 1)
    for (i = 0; i < SphereNo; i++)
      SphereA[i].rad += SolventRad;

  if (SurfDots) {
    SurfaceDots(SphereA, SphereNo, SurfKind == 2 ? SolventRad : 0.0,
	ParGetIntVal(PN_DRAW_PREC), SurfFast,
	&patchA, &patchNo);
    primP = PrimNew(PT_DOT_SURFACE, molP);
    for (i = 0; i < patchNo; i++)
      if (patchA[i].pointNo > 0)
	PrimDotSurfaceAddPatch(primP,
	    patchA[i].userData, patchA[i].pointA, patchA[i].pointNo);
    free(patchA);
  } else {
    SurfaceSolid(SphereA, SphereNo, SurfKind == 2 ? SolventRad : 0.0,
        ParGetIntVal(PN_DRAW_PREC),
	&surfA, &surfNo);
    for (surfI = 0; surfI < surfNo; surfI++) {
      primP = PrimNew(PT_SURFACE, molP);
      PrimSurfaceSetPoints(primP,
	  surfA[surfI].xA, surfA[surfI].nvA, surfA[surfI].pointNo);
      for (i = 0; i < surfA[surfI].meshNo; i++)
	PrimSurfaceAddMesh(primP,
	  surfA[surfI].meshA[i].indA, surfA[surfI].meshA[i].pointNo);
      free(surfA[surfI].meshA);
    }
    free(surfA);
  }

  free(SphereA);
}

#define ARG_NUM 5

ErrCode
ExAddSurface(char *cmd)
{
  ArgDescr arg[ARG_NUM];
  EnumEntryDescr enum1Entry[3];
  EnumEntryDescr enum2Entry[3];
  EnumEntryDescr enum3Entry[2];
  ErrCode errCode;

  arg[0].type = AT_DOUBLE;
  arg[1].type = AT_ENUM;
  arg[2].type = AT_DOUBLE;
  arg[3].type = AT_ENUM;
  arg[4].type = AT_ENUM;

  ArgInit(arg, ARG_NUM);

  arg[0].prompt = "Atom Radius";
  arg[0].entType = DE_ATOM;
  DStrAssignStr(arg[0].v.strVal, "vdw");

  enum1Entry[0].str = "vdw";
  enum1Entry[0].onOff = FALSE;
  enum1Entry[1].str = "solvent";
  enum1Entry[1].onOff = FALSE;
  enum1Entry[2].str = "contact";
  enum1Entry[2].onOff = FALSE;

  enum1Entry[SurfKind].onOff = TRUE;

  arg[1].prompt = "Surface Kind";
  arg[1].u.enumD.entryP = enum1Entry;
  arg[1].u.enumD.n = 3;
  arg[1].v.intVal = SurfKind;

  arg[2].prompt = "Solvent Radius";
  arg[2].v.doubleVal = SolventRad;

  enum2Entry[0].str = "dots";
  enum2Entry[0].onOff = SurfDots;
  enum2Entry[1].str = "shaded";
  enum2Entry[1].onOff = ! SurfDots;

  arg[3].prompt = "Surface Style";
  arg[3].u.enumD.entryP = enum2Entry;
  arg[3].u.enumD.n = 2;
  if (SurfDots)
    arg[3].v.intVal = 0;
  else
    arg[3].v.intVal = 1;

  enum3Entry[0].str = "fast";
  enum3Entry[0].onOff = SurfFast;
  enum3Entry[1].str = "safe";
  enum3Entry[1].onOff = ! SurfFast;

  arg[4].prompt = "Method";
  arg[4].u.enumD.entryP = enum3Entry;
  arg[4].u.enumD.n = 2;
  if (SurfFast)
    arg[4].v.intVal = 0;
  else
    arg[4].v.intVal = 1;

  errCode = ArgGet(arg, ARG_NUM);
  if (errCode != EC_OK) {
    ArgCleanup(arg, ARG_NUM);
    return errCode;
  }

  SurfKind = arg[1].v.intVal;
  SolventRad = arg[2].v.doubleVal;
  SurfDots = (arg[3].v.intVal == 0);
  SurfFast = (arg[4].v.intVal == 0);

  if (! SurfDots)
    /* do not allow grouping for dot surfaces because atom
       pointers are stored in the primitive there */
    DhActivateGroups(TRUE);

  BreakActivate(TRUE);

  DhApplyMol(PropGetRef(PROP_SELECTED, FALSE), addSurface, arg[0].v.exprP);

  BreakActivate(FALSE);

  if (! SurfDots)
    DhActivateGroups(FALSE);

  ArgCleanup(arg, ARG_NUM);
  GraphMolChanged(PROP_SELECTED);
  GraphRedrawNeeded();

  return EC_OK;
}
